#include "ImageGray.h"
#include "XFile.h"
#include "DiskXFile.h"

namespace Torch {

ImageGray::ImageGray()
{
	n_planes = 1;
	coding = "gray";
}

ImageGray::ImageGray(const char *filename_)
{
	n_planes = 1;
	coding = "gray";

	load(filename_);
}

ImageGray::ImageGray(int width_, int height_) : Image(width_, height_, 1)
{
	coding = "gray";
}
	
bool ImageGray::copyFrom(Image *image_)
{
	return copyFrom(image_->width, image_->height, image_->pixmap, image_->coding);
}

bool ImageGray::copyFrom(int width_, int height_, unsigned char *pixmap_, const char *coding_)
{
   	width = width_;
	height = height_;
	image_size = width * height;
	frame_size = image_size;

	imageRealloc();
	
	if(strcmp(coding_, "gray") == 0)
	{
	   	if(verbose) message("ImageGray::copyFrom(uchar *) copy from gray");
		
	   	for(int i = 0 ; i < image_size ; i++) pixmap[i] = pixmap_[i];

		updateDataFromPixmap();

		return true;
	}
	else if(strcmp(coding_, "rgb") == 0)
	{
	   	if(verbose) message("ImageGray::copyFrom(uchar *) convert from rgb");
		
	   	for(int i = 0 ; i < image_size ; i++)
		{
		   	int offset = 3 * i;
			
			pixmap[i] = rgb_to_gray(pixmap_[offset], pixmap_[offset+1], pixmap_[offset+2]);
		}

		updateDataFromPixmap();

		return true;
	}
	else if(strcmp(coding_, "yuv") == 0)
	{
	   	if(verbose) message("ImageGray::copyFrom(uchar *) convert from yuv (not implemented yet !)");
		
		updateDataFromPixmap();

		return true;
	}

   	return false;
}

bool ImageGray::copyFrom(int width_, int height_, real *data_, const char *coding_, real max_)
{
   	width = width_;
	height = height_;
	image_size = width * height;
	frame_size = image_size;

	imageRealloc();

	real rescale = 255.0 / max_;
	
	if(strcmp(coding_, "gray") == 0)
	{
	   	if(verbose) message("ImageGray::copyFrom(real *) copy from gray");
		
	   	for(int i = 0 ; i < image_size ; i++) data[i] = data_[i] * rescale;

		updatePixmapFromData();

		return true;
	}
	else if(strcmp(coding_, "rgb") == 0)
	{
	   	if(verbose) message("ImageGray::copyFrom(real *) convert from rgb");
		
	   	for(int i = 0 ; i < image_size ; i++)
		{
		   	int offset = 3 * i;
			
			data[i] = rgb_to_gray(data_[offset] * rescale, data_[offset+1] * rescale, data_[offset+2] * rescale);
		}

		updatePixmapFromData();

		return true;
	}
	else if(strcmp(coding_, "yuv") == 0)
	{
	   	if(verbose) message("ImageGray::copyFrom(real *) convert from yuv (not implemented yet !)");
		
		updatePixmapFromData();

		return true;
	}
	else if(strcmp(coding_, "float") == 0)
	{
	   	if(verbose) message("ImageGray::copyFrom(uchar *) copy from float");
		
	   	for(int i = 0 ; i < image_size ; i++) 
		{
			data[i] = data_[i] * rescale;
		}

		updatePixmapFromData();

		return true;
	}

   	return false;
}

void ImageGray::saveXFile(XFile *file)
{
	// Save as pgm (P5 format)

	file->printf("P5\n");
	file->printf("#\n");
	file->printf("# Image generated using Torch vision\n");
	file->printf("# (c) 2004-2005 Sebastien Marcel marcel@idiap.ch\n");
	file->printf("# IDIAP Research Institute\n");
	file->printf("#\n");
	file->printf("%d %d\n", width, height);
	file->printf("255\n");
   
	file->write(pixmap, sizeof(unsigned char), width*height);
}

void ImageGray::loadXFile(XFile *file)
{
	// Load as pgm (P5 format)
	
	char buffer[200];

   	// Reads Magic Number
	if(!file->gets(buffer, 200)) error("ImageGray::loadXFile() not enable to read pgm file\n");
	if(strcmp(buffer, "P5\n") != 0) error("ImageGray::loadXFile() incorrect magic number\n");

	// Reads comments
	if(!file->gets(buffer, 200)) error("ImageGray::loadXFile() not enable to read pgm file\n");
	//while(buffer[0] == '#') if(!file->gets(buffer, 200)) error("ImageGray::loadXFile() not enable to read pgm file\n");
	while(buffer[0] == '#' || buffer[0] == '\n') if(!file->gets(buffer, 200)) error("ImageGray::loadXFile() not enable to read pgm file\n");

	//
	// Reads header
   	sscanf(buffer, "%d %d\n", &width, &height);
   	if(width <= 0) error("ImageGray::loadXFile() incorrect width in pgm file\n");
   	if(height <= 0) error("ImageGray::loadXFile() incorrect height in pgm file\n");

	image_size = width * height;
	frame_size = image_size;

	//
	// Reads resolution
	if(!file->gets(buffer, 200)) error("ImageGray::loadXFile() not enable to read pgm file\n");
	if(strcmp(buffer, "255\n") != 0) error("ImageGray::loadXFile() incorrect resolution, only 255 is supported\n");

	//
	// Allocate/Reallocate memory if needed
	imageRealloc();
	
	//
	// Reads the pixmap
	int n_ = file->read(pixmap, sizeof(unsigned char), image_size);
	if(n_ != image_size) error("ImageGray::loadXFile() impossible to read the image (%d bytes read != %d bytes expected)\n", n_, image_size);

	//
	// Update image data
	updateDataFromPixmap();
}

void ImageGray::drawpixel(int x, int y, Color c)
{
	if(x>=0 && y>=0 && x<width && y<height)
	{
		/// Warning the color should be appropriate
	
		int offset = y * width + x;

		pixmap[offset] = c.data0;

		data[offset] = c.data0;
	}
}

ImageGray::~ImageGray()
{
}

}

