#ifndef IPDCTMOD2_INC
#define IPDCTMOD2_INC

#include "ipCore.h"
#include "ipDCT2D.h"

namespace Torch {

/** This class is designed to decompose the image into blocks and compute the DCTmod2 for each block

    This class computes DCTmod2 over $n \times n$ images of size (where $n = 8 \dots 64$).
    
    \begin{verbatim}
                                +--------------------------+
                                |                          |                       
                                |        ipDCTmod2         |
                                |                          |
                                |                          |
                                |   +--+      +--------+   |
				|   |AA|  --> |ipDCT2D | --|--> seq_out->frames[0]
				|   |AA|      +--------+   |
	                        |   +--+                   |
			        |			   |	  
       seq_in->frames[0]        |   +--+      +--------+   |
   	           +----+       |   |BB|  --> |ipDCT2D | --|--> seq_out->frames[1]
		   |AABB|       |   |BB|      +--------+   |
	           |AABB|  ---> |   +--+                   |
                   |CCDD|       |                          |
                   |CCDD|       |   +--+      +--------+   |
                   +----+       |   |CC|  --> |ipDCT2D | --|--> seq_out->frames[2]
                                |   |CC|      +--------+   |
                                |   +--+                   |
                                |                          |
                                |   +--+      +--------+   |
                                |   |DD|  --> |ipDCT2D | --|--> seq_out->frames[3]
                                |   |DD|      +--------+   |
                                |   +--+                   |
                                |                          |
                                +--------------------------+
    \end{verbatim}
	 
    @author Sebastien Marcel (marcel@idiap.ch)
    @author Yann Rodriguez   (rodrig@idiap.ch)
    @author Conrad Sanderson (conradsand@ieee.org)
    
    @version 2.0
    \Date
    @since 2.0

    @see ipDCT2D
    @see ipBlock
*/	
class ipDCTmod2 : public ipCore
{
   	// width of the input image
	int width;

	// height of the input image
	int height;

	// internal tmp variables
	int pixeladv;
        int row_offset;
        int col_offset;

	// subimage
	Sequence *subimage;	
	
	// internal tmp memory
	real *tmp;
	real *dct;
	real *coldelta;
	real *rowdelta;

	// number of rows (lines)
	int rowblocks_in;

	// number of columns
	int colblocks_in;

	// number of blocks
	int n_block_in;

	// computes internal variables
	void init_(void);

public:
	/// number of DCTmod2 features per block
	int output_size;

	/// number of blocks used for DCTmod2
	int n_block;

	/// number of rows (lines) used for DCTmod2
	int rowblocks;

	/// number of columns used for DCTmod2
        int colblocks;
	
	/// DCT2D used to process each block
	ipDCT2D *ip;

	/// output size of ipDCT2D and number of DCT coefficients
        int ip_output_size;

	/// block size (square)
	int block_size;

	/// overlap between block [0...block-1]
	int block_overlap;	
	
	/// number of neighbouring blocks to consider for deltas
	int delta_neighbour;

	/// number of delta coefficients to keep
	int delta_keep;

	//-----
	
	/** constructor
	
	    @param width_ is the #width# of the input image
	    @param height_ is the #height# of the input image
	    @param coding_ is the #coding# format of the input image
	    @param dim_dct is the number of DCT coefficients to keep
	    @param block_size_ is the size of blocks
	    @param block_overlap_ is the overlap between blocks
	    @param delta_neighbour_ is the number of neighbour blocks to take into accounts
	    @param delta_keep_ is the number of delta coefficients to keep
	*/
	ipDCTmod2(int width_, int height_, const char *coding_, int dim_dct, int block_size_, int block_overlap_, int delta_neighbour_, int delta_keep_);

	///
	virtual ~ipDCTmod2();

	///
	virtual void process(Sequence *seq_in);

	//-----
	
	/// get the number of rows
	int getrowblocks() { return rowblocks; };

	/// get the number of columns
	int getcolblocks() { return colblocks; };
};

}

#endif
