#ifndef IPRELAXATION_INC
#define IPRELAXATION_INC

#include "ipCore.h"
#include "Mat.h"
#include "Vec.h"
#include "multigrid.h"
#include "mx_lu_factor.h"
#include "math.h"

#define IS_NEAR(var, value, delta) ((var >= (value - delta)) && (var <= (value + delta)))

namespace Torch {

/** This class is designed to apply relaxation on the linear system induced by
    the discretization of an elliptic PDE (diffusion)

    Relaxation is an iterative method allowing the resolution (approximation)
    of large and sparse linear systems. Here, the Gauss-Seidel scheme with 
    red-black ordering is used (see multigrid.h file). The number of relaxation 
    steps can be provided (default 10).

    @author Guillaume Heusch (heusch@idiap.ch)
    @version 2.0
    \Date
    @since 2.0
*/

class ipRelaxation : public ipCore
{
  int type, steps, width, height;
  real lambda;

  real *rho;
  Vec *image;
  Vec *light;

  //---

 public:


  /** constructor
	    
	    @param lambda_ relative importance of the smoothness constraint
	    @param type_ type of diffusion (isotropic, anisotropic)
	    @param steps_ number of relaxation steps to approximate the solution
	    @param width_ is the width of the input image
	    @param height_ is the height of the input image
	    @param coding_ is the coding format of the input image
  */
  ipRelaxation(real lambda_, int type_, int steps_, int width_, int height_, const char *coding_);

  /// 
  void process(Sequence *seq_in);

  /** This function truncates the pixel's distribution of an image:
      pixels whose value is below (or above) the specified distribution width 
      are set to the new minimum (maximum) value (also defined by the 
      specified distribution width). Used for display purpose.

      @param data the data the truncate (image)
      @param width is the width of the input image
      @param height is the height of the input image
      @param distribution_width the new half-width of the pixels distribution: 
      the new distribution is in the interval [mean - distribution_width*std_dev; mean + distribution_width*std_dev]

   */
  void cutExtremum(real *data, int width, int height, int distribution_width);

  ///
  virtual ~ipRelaxation();
};

}
#endif  

  

