#ifndef IPVCYCLE_INC
#define IPVCYCLE_INC

#include "ipCore.h"
#include "Mat.h"
#include "Vec.h"
#include "multigrid.h"
#include "mx_lu_factor.h"
#include "math.h"

#define IS_NEAR(var, value, delta) ((var >= (value - delta)) && (var <= (value + delta)))

namespace Torch {

/** This class implements the multigrid V-cycle algorithm

    The V-cycle algorithm is an iterative method allowing the resolution (approximation)
    of large and sparse linear systems induced by partial differential equations. 
    Multiple grids of different resolution are used in order to speed up the resolution. 
    Note that the number of grids is dependent on the initial size of the 2D data.
   
    \begin{verbatim}
    "A Multigrid Tutorial", W. L. Briggs. SIAM Books, Philadelphia, 1987.
    \end{verbatim}

    @author Guillaume Heusch (heusch@idiap.ch)
    @version 2.0
    \Date
    @since 2.0
*/
class ipVcycle : public ipCore
{
  int width, height;
  int n_grids, type;
  real *rho;
  real lambda;
  
  Vec *image_grid; 
  Vec *light;
  
  Mat *diffOperator;
	
  //---

 public:	

  /** constructor

  @param lambda relative importance of the smoothness constraint
  @param n_grids number of grids used in the v-cycle (dependent on image size)
  @param width_ is the width of the input image
  @param height_ is the height of the input image
  @param coding_ is the coding format of the input image
   */
  ipVcycle(real lambda, int n_grids, int type, int width, int height, const char *coding);
  
  ///
  void process(Sequence *seq_in);
  
  /** this function implements the v-cycle (recursive call), in order to solve 
      the large and sparse linear system Ax=b
      
  @param x the result
  @param b the right-hand side term in Ax=b
  @param lambda relative importance of the smoothness constraint
  @param level 'depth' of the current grid (0 = finest grid)
  @param type type of diffusion (isotropic, anisotropic)
   */
  Vec* mgv(Vec *x, Vec *b, real lambda, int level, int type);
  
  void cutExtremum(real *data, int width, int height, int distribution_width);
    
  virtual ~ipVcycle();
};
 
}

#endif
