#include "HaarFeatureMaker.h"
#include "XFile.h"

namespace Torch {

/**------------------- HAARFEATUREMAKER ------------------
*/
HaarFeatureMaker::HaarFeatureMaker(int n_mask_id_, int width_, int height_, int min_size_, int max_size_)
{
	if((n_mask_id_ <= 0) || (n_mask_id_ > 5))
		error("HaarFeatureMaker doesn't support this range of mask_ids (%d)", n_mask_id_);

	// ouais...
	n_mask_id = n_mask_id_;

	width = width_;
	height = height_;

	if(min_size_ < 2)
	{
		warning("min_size must be >= 2 (=> set to 2)");
		min_size = 2;
	}
	if(min_size_%2 != 0) min_size = min_size_ + 1; // valeur paire
	else min_size = min_size_;
	max_size = max_size_;

	// first, find the size of the feature space
	n_features = 0;
	for(int i = 1; i <= n_mask_id; i++)
	{
		if(i == 1) computeFeatures(2, 1, i, false);
		if(i == 2) computeFeatures(1, 2, i, false);
		if(i == 3) computeFeatures(2, 2, i, false);
		if(i == 4) computeFeatures(3, 1, i, false);
		if(i == 5) computeFeatures(1, 3, i, false);
	}
	message("HaarFeatureMaker: size of the feature space: %d", n_features);

	mask = (ipHaar **)allocator->alloc(sizeof(ipHaar*)*n_features);
	for(int i = 0 ; i < n_features ; i++) mask[i] = NULL;

	// and then create the feature vector
	n_features = 0;
	for(int i = 1; i<= n_mask_id; i++)
	{
		if(i == 1) computeFeatures(2, 1, i, true);
		if(i == 2) computeFeatures(1, 2, i, true);
		if(i == 3) computeFeatures(2, 2, i, true);
		if(i == 4) computeFeatures(3, 1, i, true);
		if(i == 5) computeFeatures(1, 3, i, true);
	}
}

void HaarFeatureMaker::computeFeatures(int w_step, int h_step, int mask_id_, bool create_)
{
	int shift;
	int min_size_w = min_size;
	int min_size_h = min_size;
	
	switch(mask_id_)
	{
	 case 1: // Bi-Vertical 
	 case 2: // Bi-Horizontal 
         case 3: // Diagonal 
		 break;
	 case 4: // Tri-Vertical 
		 if(min_size%3 != 0) min_size_w += 3 - (min_size%3); // min_size_w doit etre un multiple de trois
		 break;
	 case 5: // Tri-Horizontal 
		 if(min_size%3 != 0) min_size_h += 3 - (min_size%3); // min_size_h doit etre un multiple de trois
		 break;
	}

	for(int mask_size_w = min_size_w ; mask_size_w <= max_size ; mask_size_w +=w_step)
	{
		for(int mask_size_h = min_size_h ; mask_size_h <= max_size ; mask_size_h +=h_step)
		{
			if(mask_size_h > mask_size_w) shift = mask_size_w / 4;
			else shift = mask_size_h / 4;
			if(shift == 0) shift = 1;
			
			int w_max = width - mask_size_w;
			int h_max = height - mask_size_h;
			
			for(int x_pos = 1 ; x_pos <= w_max ; x_pos += shift)
				for(int y_pos = 1 ; y_pos <= h_max ; y_pos += shift)
				{
					// allocate and create the feature vector of Haar-Like masks
					if(create_ == true)
					{
						switch(mask_id_)
						{
						 case 1: /* Bi-Vertical */
							 mask[n_features] = (ipHaar *) new(allocator) biVertHaar(width, height, x_pos, y_pos, mask_size_w, mask_size_h);
							 break;
						 case 2: /* Bi-Horizontal */
							 mask[n_features] = (ipHaar *) new(allocator) biHorizHaar(width, height, x_pos, y_pos, mask_size_w, mask_size_h);
							 break;
						 case 3: /* Diagonal */
							 mask[n_features] = (ipHaar *) new(allocator) diagHaar(width, height, x_pos, y_pos, mask_size_w, mask_size_h);
							 break;
						 case 4: /* Tri-Vertical */
							 mask[n_features] = (ipHaar *) new(allocator) triVertHaar(width, height, x_pos, y_pos, mask_size_w, mask_size_h);
							 break;
						 case 5: /* Tri-Horizontal */
							 mask[n_features] = (ipHaar *) new(allocator) triHorizHaar(width, height, x_pos, y_pos, mask_size_w, mask_size_h);
							 break;
						}
					}
					n_features++;
				}
		}
	}
}

HaarFeatureMaker::~HaarFeatureMaker()
{
}

}




