#include "VideoGrabberV4L.h"
#include "Convert.h"

namespace Torch {

//---
//
static int NbVideoAttributes = 5;
static char *VideoAttributes[] = {	"brightness", 
					"contrast", 
					"hue", 
					"color",
					"gamma"};


//
static unsigned int MinMax_VideoAttributes_[15] = {	0, 65535, 40195,
							0, 65535, 27524,
							0, 65535, 32768,
							0, 65535, 32512,
							0, 65535, 32512};

struct sMinMax_VideoAttribute_ { int min_va, max_va, default_va; };
static sMinMax_VideoAttribute_ *MinMax_VideoAttributes = (sMinMax_VideoAttribute_ *) MinMax_VideoAttributes_;

//---
//
#ifdef HAVE_V4L
static int NbVideoNorm = 6;
static char *VideoNorm[] = {	"PAL", 
				"NTSC", 
				"SECAM", 
				"AUTO", 
				"Progressive Scan PAL",
				"Progressive Scan NTSC"};
#endif


//
struct sVideoFormat_ { char *video_format; int v4l_video_format_id; };
#ifdef HAVE_V4L
static int NbNativeVideoFormats = 16;
static sVideoFormat_ NativeVideoFormats[] = {
						{"GREY", VIDEO_PALETTE_GREY},
						{"HI240", VIDEO_PALETTE_HI240},
						{"RGB565", VIDEO_PALETTE_RGB565},
						{"RGB24", VIDEO_PALETTE_RGB24},
						{"RGB32", VIDEO_PALETTE_RGB32},
						{"RGB555", VIDEO_PALETTE_RGB555},
						{"YUV422", VIDEO_PALETTE_YUV422},
						{"YUYV", VIDEO_PALETTE_YUYV},
						{"UYVY", VIDEO_PALETTE_UYVY},
						{"YUV420", VIDEO_PALETTE_YUV420},
						{"YUV411", VIDEO_PALETTE_YUV411},
						{"RAW", VIDEO_PALETTE_RAW},
						{"YUV422P", VIDEO_PALETTE_YUV422P},
						{"YUV411P", VIDEO_PALETTE_YUV411P},
						{"YUV420P", VIDEO_PALETTE_YUV420P},
						{"YUV410P", VIDEO_PALETTE_YUV410P}};
#else
static int NbNativeVideoFormats = 0;
static sVideoFormat_ NativeVideoFormats[] = {};
#endif

//---
// List of functions to convert from YUV422 to: yuv, rgb
struct VideoGrabber_output_format_encode_ VideoGrabberYUV422P_output_format_encode[] = {
						{"YUV", YUV422P_to_YUV},
						{"RGB24", YUV422P_to_RGB24}};

//---
// List of functions to convert from YUV420P to: gray, yuv and rgb
struct VideoGrabber_output_format_encode_ VideoGrabberYUV420P_output_format_encode[] = {
						{"GREY", YUV420P_to_GREY},
						{"YUV", YUV420P_to_YUV},
						{"RGB24", YUV420P_to_RGB24}};

//---
// List of functions to convert from RGB24 to: rgb
struct VideoGrabber_output_format_encode_ VideoGrabberRGB24_output_format_encode[] = {
						{"RGB24", RGB24_to_RGB24}};

//---
// List of functions to convert from
static int NbVideoFormatDecode = 17;	 

struct VideoGrabber_video_format_decode_ VideoGrabberV4L_video_format_decode[] = {
						{"GREY",	8, 0, NULL},
						{"HI240",	8, 0, NULL},
						{"RGB565",	16, 0, NULL},
						{"RGB24",	24, 1, VideoGrabberRGB24_output_format_encode},
						{"RGB32",	32, 0, NULL},
						{"RGB555",	16, 0, NULL},
						{"YUV422",	16, 0, NULL},
						{"YUYV",	32, 0, NULL},
						{"UYVY",	32, 0, NULL},
						{"YUV420",	24, 0, NULL},
						{"YUV411",	24, 0, NULL},
						{"RAW",		24, 0, NULL},
						{"YUV",         24, 0, NULL}, /* yuvyuvyuvyuv */
						{"YUV422P",	16, 2, VideoGrabberYUV422P_output_format_encode},
						{"YUV411P",	24, 0, NULL},
						{"YUV420P",	12, 3, VideoGrabberYUV420P_output_format_encode},
						{"YUV410P",	12, 0, NULL}};

/** FrameV4L methods.
*/

FrameV4L::FrameV4L(int width_, int height_, int input_video_format_, int output_video_format_, bool verbose_)
{
   	verbose = verbose_;

	if((input_video_format_ != -1) && (output_video_format_ != -1))
	{
		//
		width = width_;
   		height = height_;


		//
		input_video_format = input_video_format_;

        	bytesperline_in = width * VideoGrabberV4L_video_format_decode[input_video_format].bytesperline / 8;
       
		if(verbose == true)
		{
        		print("Input %s(%d):\n", VideoGrabberV4L_video_format_decode[input_video_format].image_coding, input_video_format);
			print("  w = %d\n", width);
			print("  h = %d\n", height);
        		print("  bpl = %d\n", bytesperline_in);
	        	print("  size = %d\n", height * bytesperline_in);
		}

		//
		output_video_format = output_video_format_;
		
		//
		// Match output video format into VideoGrabberV4L_video_format_decode to find bytesperline
		if(verbose == true)
			print("Matching output video format %s(%d) to decoding look-up table:\n", VideoGrabberV4L_video_format_decode[input_video_format].output_format_encode[output_video_format].image_coding, output_video_format);

		int video_format_ = -1;
		for(int i = 0 ; i < NbVideoFormatDecode ; i++)
		{
			if(verbose == true)
				print("   Trying %s ...\n", VideoGrabberV4L_video_format_decode[i].image_coding);

			if(strcmp(VideoGrabberV4L_video_format_decode[input_video_format].output_format_encode[output_video_format].image_coding, VideoGrabberV4L_video_format_decode[i].image_coding) == 0)
			{
				video_format_ = i;
			}
		}

		if(verbose == true)
			printf("Match found %s(%d) == %s(%d)\n",
						VideoGrabberV4L_video_format_decode[input_video_format].output_format_encode[output_video_format].image_coding, 
						output_video_format,	 
						VideoGrabberV4L_video_format_decode[video_format_].image_coding, 
						video_format_);

        	bytesperline_out = width * VideoGrabberV4L_video_format_decode[video_format_].bytesperline / 8;
		int SizeOutput = height * bytesperline_out;

		if(verbose == true)
		{
        		print("Output %s(%d):\n", VideoGrabberV4L_video_format_decode[input_video_format].output_format_encode[output_video_format].image_coding, output_video_format);
			print("  w = %d\n", width);
			print("  h = %d\n", height);
        		print("  bpl = %d\n", bytesperline_out);
	        	print("  size = %d\n", SizeOutput);
		}

		//
  		data = NULL;  

  		data = new unsigned char [SizeOutput];

		//
  		bIsUsed = false;

#ifdef FAST_ICCIRYUV_RGB
		initLUT_ICCIRYUV_RGB();
#endif
	}
	else
	{
		data = NULL;

		warning("FrameV4L: Something goes wrong with input/output video formats (%d/%d)", input_video_format_, output_video_format_);
	}
}

FrameV4L::~FrameV4L()
{
  	if(data != NULL) delete [] data;
}

unsigned int FrameV4L::getwidth() const
{
	return width;
}

unsigned int FrameV4L::getheight() const
{
  	return height;
}

const unsigned char *FrameV4L::getpixmap() const
{
  	return data;
}

void FrameV4L::copyBuffer(unsigned char *v4l_buffer)
{
	// call appropriate video converter
	VideoGrabberV4L_video_format_decode[input_video_format].output_format_encode[output_video_format].video_decode_output_encode(
							width, height,
     							(unsigned char *) v4l_buffer, bytesperline_in,  
   							(unsigned char *) data, bytesperline_out);
/*
	if(data != NULL)
	{
		if(VideoGrabberV4L_video_format_decode[input_video_format].output_format_encode != NULL)
		{
	   		if(VideoGrabberV4L_video_format_decode[input_video_format].output_format_encode[output_video_format].video_decode_output_encode != NULL)  
			{
				// call appropriate video converter
				VideoGrabberV4L_video_format_decode[input_video_format].output_format_encode[output_video_format].video_decode_output_encode(
											width, height,
		      									(unsigned char *) v4l_buffer, bytesperline_in,  
		      									(unsigned char *) data, bytesperline_out);
			}
		}
	}
*/
}

/** VideoGrabberV4L methods
*/

VideoGrabberV4L::VideoGrabberV4L(	const char *devname_,
					int width_, int height_, 
				 	const char *output_video_format_,
				 	const char *input_video_format_,
				 	int n_buffers_,
				 	int video_channel_,
					int frame_rate_,
					bool verbose_) : VideoGrabber(verbose_)
{
#ifndef HAVE_V4L
	error("HAVE_V4L not defined in VideoGrabberV4L::VideoGrabberV4L()");
#endif
   
   	//
	strcpy(devname, (char *) devname_);
	
	//
  	width = width_;
  	height = height_;

	//
  	n_buffers = n_buffers_;
  	video_channel = video_channel_;
	frame_rate = frame_rate_;
  
	//
	status = VIDEOSOURCE_OK;

  	bool bOkVideoFormatFound;  
 
	// 
	// Match the input video format with the list of native V4L video formats
	if(verbose == true)
		print("Matching native video format %s:\n", input_video_format_);
	bOkVideoFormatFound = false;  
  	for(int i = 0 ; i < NbNativeVideoFormats ; i++)
	{
		if(verbose == true)
			print("   Trying %s ...\n", NativeVideoFormats[i].video_format);

		if(strcmp(input_video_format_, NativeVideoFormats[i].video_format) == 0)
		{
			bOkVideoFormatFound = true;
			v4l_native_video_format = i;
			 
			break;
		}								  
	}

	if(bOkVideoFormatFound == true)
	{
		if(verbose == true)
			print("Match found %s -> %d\n\n", input_video_format_, v4l_native_video_format);

		//
		// Match native video format into VideoGrabberV4L_video_format_decode
		if(verbose == true)
			print("Matching video format %s(%d) to decoding look-up table:\n", NativeVideoFormats[v4l_native_video_format].video_format, v4l_native_video_format);
		int video_format = -1;
		bOkVideoFormatFound = false;  
		for(int i = 0 ; i < NbVideoFormatDecode ; i++)
		{
			if(verbose == true)
				print("   Trying %s ...\n", VideoGrabberV4L_video_format_decode[i].image_coding);

			if(strcmp(NativeVideoFormats[v4l_native_video_format].video_format, VideoGrabberV4L_video_format_decode[i].image_coding) == 0)
			{
				bOkVideoFormatFound = true;
				video_format = i;
			}
		}

		if(bOkVideoFormatFound == true)
		{
		   	input_video_format = video_format;
			
			//
			// Match the output video format with the list of VideoGrabberV4L output video formats
			if(verbose == true)
				print("Matching output video format %s to decoding look-up table %s(%d):\n", 
								output_video_format_, 
								VideoGrabberV4L_video_format_decode[input_video_format].image_coding, input_video_format);
			bOkVideoFormatFound = false;  
	  		for(int i = 0 ; i < VideoGrabberV4L_video_format_decode[input_video_format].NbOutputFormatEncode ; i++)
			{
				if(verbose == true)
					print("   Trying %s ...\n", VideoGrabberV4L_video_format_decode[input_video_format].output_format_encode[i].image_coding);

				if(strcmp(output_video_format_, VideoGrabberV4L_video_format_decode[input_video_format].output_format_encode[i].image_coding) == 0)
				{
					bOkVideoFormatFound = true;
					output_video_format = i;
			 
					break;
				}								  
			}
		
			if(bOkVideoFormatFound == true)
			{
				if(verbose == true)
					print("Match found %s -> %d\n\n", output_video_format_, output_video_format);
  			}
			else
			{		
      				warning("VideoGrabberV4L: Unknowned output video format %s.", output_video_format_);

				if(verbose == true)
				{
					print("  Supported Output Video Format for %s Video Format are : [ ", input_video_format_);
		
					for(int i = 0 ; i < VideoGrabberV4L_video_format_decode[input_video_format].NbOutputFormatEncode ; i++)
   						print("%s ", VideoGrabberV4L_video_format_decode[input_video_format].output_format_encode[i].image_coding);
					print("]\n");		
				}

	      			status = VIDEOSOURCE_ERROR;
			}
		}
  		else	
		{		
      			warning("VideoGrabberV4L: Unknowned input video format %s.", input_video_format_);

			if(verbose == true)
			{
				print("  Supported Transcoding Video Format are : [ ");
		
				for(int i = 0 ; i < NbVideoFormatDecode ; i++)
   					print("%s ", VideoGrabberV4L_video_format_decode[i].image_coding);
				print("]\n");		
			}

      			status = VIDEOSOURCE_ERROR;
		}
	}
	else
	{		
      		warning("VideoGrabberV4L: Unknowned native V4L video format %s.", input_video_format_);

		if(verbose == true)
		{
			print("  Supported Video Format are : [ ");
		
			for(int i = 0 ; i < NbNativeVideoFormats ; i++)
   				print("%s ", NativeVideoFormats[i].video_format);
			print("]\n");		
		}

      		status = VIDEOSOURCE_ERROR;
	}

	camfd = -1;
}

bool VideoGrabberV4L::initVideoSource()
{
  	if(status == VIDEOSOURCE_ERROR) return false;  

  	if(width%4 != 0)
	{
		warning("VideoGrabberV4L: the width should be a multiple of 4.");
		
		status = VIDEOSOURCE_ERROR;

		return false;		
	}

	if(verbose == true)
	{
		print("Video Attributes:\n");
		for(int va = 0 ; va < NbVideoAttributes ; va++)
			print(" %s: [min=%d, max=%d, default=%d\n", VideoAttributes[va], MinMax_VideoAttributes[va].min_va, MinMax_VideoAttributes[va].max_va, MinMax_VideoAttributes[va].default_va); 
	}

	return true; 
}

bool VideoGrabberV4L::openVideoSource()
{
#ifdef HAVE_V4L
	if(verbose == true)
		print("VideoGrabberV4L::Open()\n");

  	if(status == VIDEOSOURCE_OK)
	{
		camfd = open(devname, O_RDONLY);

		if(camfd < 0)
		{
			warning("VideoGrabberV4L: No grabber device available");

      			status = VIDEOSOURCE_ERROR;
	
			return false;
    		}
	
		if(status == VIDEOSOURCE_OK)
		{
    			if(ioctl(camfd, VIDIOCGCAP, &vcap) < 0)
			{
				warning("VideoGrabberV4L: Could not query capabilities");

	      			status = VIDEOSOURCE_ERROR;
	
				return false;
    			}
			else vcap.name[31] = '\0';

			vchan.channel = 0;
			if(ioctl(camfd, VIDIOCGCHAN, &vchan) < 0)
			{
				warning("VideoGrabberV4L: Could not query video channel");

	      			status = VIDEOSOURCE_ERROR;
	
				return false;
    			}
			else vchan.name[31] = '\0';

		} // End checking device
	
		if(status == VIDEOSOURCE_OK)
		{
			if(verbose == true)
			{
			   	print(" + name = %s\n", vcap.name);
				print(" + number of video inputs = %d\n", vcap.channels);
				print("    * max width = %d\n", vcap.maxwidth);
				print("    * max height = %d\n", vcap.maxheight);
				print("    * min width = %d\n", vcap.minwidth);
				print("    * min height = %d\n", vcap.minheight);
				print(" + number of audio inputs = %d\n", vcap.audios);
				print(" + current video channel:\n");
				print("    * name = %s (%d)\n", vchan.name, vchan.channel);

				if(vchan.norm < NbVideoNorm)
					print("    * norm = %s (%d)\n", VideoNorm[vchan.norm], vchan.norm);
				else print("    * norm = ??? (%d)\n", vchan.norm);
			}
			status = VIDEOSOURCE_OPEN;
		}
	}
  	else
	{
		warning("VideoGrabberV4L: Impossible to open (status problem).");

      		return false;
	}

	if(vchan.channel != video_channel)
	{
		if(verbose == true) print("Setting video channel...\n");
	
		vchan.channel = video_channel;
		vchan.norm = 0; // forcing to PAL
		if(ioctl(camfd, VIDIOCSCHAN, &vchan) < 0)
		{
			print("Error: could not set video channel ...\n");

			status = VIDEOSOURCE_ERROR;

			return false;
		}

		if(ioctl(camfd, VIDIOCGCHAN, &vchan) < 0)
		{
			warning("VideoGrabberV4L: Could not query video channel a second time");

   			status = VIDEOSOURCE_ERROR;
	
			return false;
    		}

		if(verbose == true)
		{
			print(" + new video channel:\n");
			print("    * name = %s (%d)\n", vchan.name, vchan.channel);
			if(vchan.norm < NbVideoNorm)
				print("    * norm = %s (%d)\n", VideoNorm[vchan.norm], vchan.norm);
			else print("    * norm = ??? (%d)\n", vchan.norm);
		}
	}

	//
	if (ioctl(camfd, VIDIOCGPICT, &VPic) < 0)
	{
		print("Error: could not get picture parameters ...\n");

		status = VIDEOSOURCE_ERROR;

		return false;
	}

	if(verbose == true)
	{
		print("Picture parameters:\n");
		print(" + brightness = %d\n", VPic.brightness);
		print(" + contrast = %d\n", VPic.contrast);
		print(" + hue = %d\n", VPic.hue);
		print(" + colour = %d\n", VPic.colour);
		print(" + gamma (whiteness) = %d\n", VPic.whiteness);
		print(" + depth = %d\n", VPic.depth);
		print(" + palette = %d\n", VPic.palette);
	}

	//
	setBrightness(MinMax_VideoAttributes[0].default_va);
	setContrast(MinMax_VideoAttributes[1].default_va);
  	setHue(MinMax_VideoAttributes[2].default_va);
	setColor(MinMax_VideoAttributes[3].default_va);  
	setGamma(MinMax_VideoAttributes[4].default_va);  

	//
	if(verbose == true)
		print("Probing palette capabilities:\n");
	bool palette_probed = false;
	int palette = NativeVideoFormats[v4l_native_video_format].v4l_video_format_id;

	for(int i = 0 ; i < NbNativeVideoFormats ; i++)
	{
		if(verbose == true)
			print(" + palette[%d] %s: ", i+1, NativeVideoFormats[i].video_format);
		   
		VPic.palette = NativeVideoFormats[i].v4l_video_format_id;

		if (ioctl(camfd, VIDIOCSPICT, &VPic) < 0)
		{
			if(verbose == true)
				print("FAILED\n");
		}
		else
		{
			if(verbose == true) print("OK\n");

			if(palette == NativeVideoFormats[i].v4l_video_format_id) palette_probed = true;
		}
	}

	if(!palette_probed)
	{
		print("Error: video format %s not supported.\n", NativeVideoFormats[v4l_native_video_format].video_format);
		return false;
	}


	//
	struct video_window vwin;

	memset(&vwin, 0, sizeof(struct video_window));

	//
	if (ioctl(camfd, VIDIOCGWIN, &vwin) < 0)
	{
		print("Error: could not get picture size ...\n");
		
		status = VIDEOSOURCE_ERROR;
		
		return false;		
	}

#if defined(PWC_FPS_SHIFT)
	if ((vwin.flags & PWC_FPS_FRMASK) >> PWC_FPS_SHIFT)
	{
		if(verbose) print("VideoDevice supports framerate setting [%x].\n", vwin.flags);

		int FrameRate = frame_rate;
		if(FrameRate <= 0) FrameRate = 0;
		if(FrameRate > 63) FrameRate = 63;

		vwin.flags = (vwin.flags & ~PWC_FPS_MASK) | (FrameRate << PWC_FPS_SHIFT);

		if(ioctl(camfd, VIDIOCSWIN, &vwin) < 0) 
		{
			print("Error: could not set frame rate ...\n");
		
			status = VIDEOSOURCE_ERROR;
		
			return false;		
		}
	}
	else print("VideoDevice does not supports framerate setting.\n");
#endif
	

	if(verbose == true)
	{
		print("Picture size:\n");
		print(" + width = %d\n", vwin.width);
		print(" + height = %d\n", vwin.height);
	}

	vm.format = NativeVideoFormats[v4l_native_video_format].v4l_video_format_id;
	vm.width = width;
	vm.height = height;

  	Frame = (FrameV4L **) malloc(sizeof(FrameV4L *) *2);
  	Frame[0] = new FrameV4L(width, height, input_video_format, output_video_format, verbose);
  	Frame[1] = new FrameV4L(width, height, input_video_format, output_video_format, verbose);
  
  	if(Frame == NULL)
	{
		warning("VideoGrabberV4L: memory allocation for Frame buffer.");
		
		status = VIDEOSOURCE_ERROR;
		
		return false;		
	}

	current_frame = 0;

	//
	current_buffer = 0;

	//
	// see if device has mmap
	VMBuf.size = 0;
	VMBuf.frames = 0;

	//
	vid_io_buffer_size = 0;

	if (ioctl(camfd, VIDIOCGMBUF, &VMBuf) == 0) 
	{
		if(verbose == true)
			print("Using mmap(): VMBuf.size = %d, VMBuf.frames = %d\n", VMBuf.size, VMBuf.frames);

		vid_io_buffer_size = VMBuf.size;

		if (n_buffers > 0) 
		{
			// User specified a number of buffers; see if we can match that.
			if (VMBuf.frames < n_buffers) 
			{
				print("Warning: there are more buffers requested than MBUF can provide. Limiting buffers.\n");
				n_buffers = VMBuf.frames;
			}
		}
		else // We grab everything we can.
			n_buffers = VMBuf.frames;
	}
	else 
	{
		VMBuf.size = 0; // Just to be sure....
		VMBuf.frames = 0;
	}
		
	if(verbose == true)
		print("n_buffers = %d\n", n_buffers);

	//
	// See if we can actually mmap the memory
	vid_io_offsets = NULL;
	if (VMBuf.size > 0)
	{
		vid_io_buffer = (unsigned char *) mmap(NULL, vid_io_buffer_size, PROT_READ, MAP_SHARED, camfd, 0);
		if (vid_io_buffer == NULL)
		{
			print("Warning: mmap() failed (%d). Falling back to non-mmaped mode.", errno);
			VMBuf.size = 0;
			vid_io_buffer = NULL;
		}
		else
		{
			vid_io_offsets = new int [n_buffers];
			for (int i = 0; i < n_buffers; i++) vid_io_offsets[i] = VMBuf.offsets[i];
		}
	}

	if (VMBuf.size == 0)
	{
		// No mmap or failed

		if(verbose == true)
			print("Allocating own buffer ...\n");

		if(n_buffers < 0)
		{
			n_buffers = 4;
			vid_io_buffer_size = n_buffers * vcap.maxwidth * vcap.maxheight * 4;
			vid_io_buffer = new unsigned char[vid_io_buffer_size];
			vid_io_offsets = new int [n_buffers];
			for (int i = 0; i < n_buffers; i++) vid_io_offsets[i] = i * vcap.maxwidth * vcap.maxheight * 4;
		}
		if (vid_io_buffer == NULL)
			print("Warning: Failed to mmap/allocate memory!\n");
	}

	Capture(0);

	return true;
#else
	return false;
#endif
}

bool VideoGrabberV4L::closeVideoSource()
{
#ifdef HAVE_V4L
  	if(status == VIDEOSOURCE_OPEN)
	{
		status = VIDEOSOURCE_OK;

		if (VMBuf.size > 0)
		{
			// releasing buffer
			if (VMBuf.size == 0) return 0;

			ioctl(camfd, VIDIOCSYNC, &current_buffer);

			//
			munmap(vid_io_buffer, vid_io_buffer_size);
			VMBuf.size = 0;
		}
		else delete [] vid_io_buffer;	
			
		if(vid_io_offsets != NULL) delete [] vid_io_offsets;

		vid_io_buffer = NULL;
		vid_io_buffer_size = 0;

		//
		close(camfd);
		camfd = -1;
	
      		return true;
	}
  	else
	{
		warning("VideoGrabberV4L: Impossible to close (status problem).");

      		return false;
	}
#else
	return false;
#endif
}

int VideoGrabberV4L::Capture(int the_buffer_)
{
#ifdef HAVE_V4L
	vm.frame = the_buffer_;

	if(verbose)
		print("Capture(): buffer %d, format %d, size (%dx%d)\n", the_buffer_, NativeVideoFormats[v4l_native_video_format].v4l_video_format_id, width, height);

	if (ioctl(camfd, VIDIOCMCAPTURE, &vm) < 0)
	{
		print("Error: Capture ioctl()\n");
		return -errno;
	}
#endif
	return 0;
}

int VideoGrabberV4L::Sync()
{
#ifdef HAVE_V4L
	if (ioctl(camfd, VIDIOCSYNC, &current_buffer) < 0) 
	{
		print("Error: Sync() ioctl\n");
		return -1;
	}
#endif 
	return 0;
}

VirtualPixmap *VideoGrabberV4L::getframe()
{
#ifdef HAVE_V4L
  	if(status != VIDEOSOURCE_OPEN)
	{
		warning("VideoGrabberV4L: Impossible to get a frame (status problem).");

      		return NULL;
	}
 
	unsigned char *src;

	src = vid_io_buffer + vid_io_offsets[current_buffer];
	if (VMBuf.size > 0) 
	{ // mmap()ed.
		if (Sync() < 0)
		{
			warning("VideoGrabberV4L: Capturing image failed.");

			return false;
		}
		if (n_buffers > 1) // we have space, start capture immediately in next buffer. Otherwise, see end of function
			Capture((current_buffer + 1) % n_buffers); 
	}
	else 
	{
		if (read(camfd, src, vid_io_image_size) < 0)
		{
			warning("VideoGrabberV4L: Capturing image failed.");

			return false;
		}
	}

	current_frame = (current_frame + 1) % 2;
  	Frame[current_frame]->copyBuffer(src);
		
	/* Go to next buffer (including capture). In case of 1 buffer current_buffer
	   will get stuck at 0, so we do this after we processed the buffer content.
	*/
	current_buffer = (current_buffer + 1) % n_buffers;
	if (n_buffers == 1) 
	{
		if (Capture(current_buffer) < 0) return NULL;
	}
	
  	return Frame[current_frame];
#else
	return NULL;
#endif
}

void VideoGrabberV4L::Flush()
{
  	if(Frame != NULL)
	{
      		delete Frame[0];
      		delete Frame[1];
		free(Frame);

      		Frame = NULL;
	}

}

int VideoGrabberV4L::getBrightness()
{
#ifdef HAVE_V4L
  	if((status != VIDEOSOURCE_OK) && (status != VIDEOSOURCE_OPEN)) return -1;

	return VPic.brightness;
#else
	return -1;
#endif
}

bool VideoGrabberV4L::setBrightness(int new_value)
{
#ifdef HAVE_V4L
  	int va = 0;
    	int previous_value = getBrightness();

  	if(previous_value < 0) return false;
  	else
	{		  
		if((new_value < MinMax_VideoAttributes[va].min_va) || (new_value > MinMax_VideoAttributes[va].max_va))
		{			 
			warning("VideoGrabberV4L: Video Attribute (%s) value %d out of range, setting to default (%d)", VideoAttributes[va], new_value, MinMax_VideoAttributes[va].default_va); 

			new_value = MinMax_VideoAttributes[va].default_va;			 
		}

		if(verbose == true) print("setting %s (%d) to %d\n", VideoAttributes[va], previous_value, new_value);

		VPic.brightness = new_value & 0xffff;
		if (ioctl(camfd, VIDIOCSPICT, &VPic) < 0) return false;
		return true;		
	}  
#else
	return false;
#endif
}

int VideoGrabberV4L::getContrast()
{
#ifdef HAVE_V4L
  	if((status != VIDEOSOURCE_OK) && (status != VIDEOSOURCE_OPEN)) return -1;

	return VPic.contrast;
#else
	return -1;
#endif
}

bool VideoGrabberV4L::setContrast(int new_value)
{
#ifdef HAVE_V4L
  	int va = 1;
    	int previous_value = getContrast();

  	if(previous_value < 0) return false;
  	else
	{		  
		if((new_value < MinMax_VideoAttributes[va].min_va) || (new_value > MinMax_VideoAttributes[va].max_va))
		{			 
			warning("VideoGrabberV4L: Video Attribute (%s) value %d out of range, setting to default (%d)", VideoAttributes[va], new_value, MinMax_VideoAttributes[va].default_va); 

			new_value = MinMax_VideoAttributes[va].default_va;			 
		}

		if(verbose == true) print("setting %s (%d) to %d\n", VideoAttributes[va], previous_value, new_value);

		VPic.contrast = new_value & 0xffff;
		if (ioctl(camfd, VIDIOCSPICT, &VPic) < 0) return false;
		return true;		
	}  
#else
	return false;
#endif
}


int VideoGrabberV4L::getHue()
{
#ifdef HAVE_V4L
  	if((status != VIDEOSOURCE_OK) && (status != VIDEOSOURCE_OPEN)) return -1;

	return VPic.hue;
#else
	return -1;
#endif
}

bool VideoGrabberV4L::setHue(int new_value)
{
#ifdef HAVE_V4L
  	int va = 2;
    	int previous_value = getHue();

  	if(previous_value < 0) return false;
  	else
	{		  
		if((new_value < MinMax_VideoAttributes[va].min_va) || (new_value > MinMax_VideoAttributes[va].max_va))
		{			 
			warning("VideoGrabberV4L: Video Attribute (%s) value %d out of range, setting to default (%d)", VideoAttributes[va], new_value, MinMax_VideoAttributes[va].default_va); 

			new_value = MinMax_VideoAttributes[va].default_va;			 
		}

		if(verbose == true) print("setting %s (%d) to %d\n", VideoAttributes[va], previous_value, new_value);

		VPic.hue = new_value & 0xffff;
		if (ioctl(camfd, VIDIOCSPICT, &VPic) < 0) return false;
		return true;		
	}  
#else
	return false;
#endif
}


int VideoGrabberV4L::getColor()
{
#ifdef HAVE_V4L
  	if((status != VIDEOSOURCE_OK) && (status != VIDEOSOURCE_OPEN)) return -1;

	return VPic.colour;
#else
	return -1;
#endif
}

bool VideoGrabberV4L::setColor(int new_value)
{
#ifdef HAVE_V4L
  	int va = 3;
    	int previous_value = getColor();

  	if(previous_value < 0) return false;
  	else
	{		  
		if((new_value < MinMax_VideoAttributes[va].min_va) || (new_value > MinMax_VideoAttributes[va].max_va))
		{			 
			warning("VideoGrabberV4L: Video Attribute (%s) value %d out of range, setting to default (%d)", VideoAttributes[va], new_value, MinMax_VideoAttributes[va].default_va); 

			new_value = MinMax_VideoAttributes[va].default_va;			 
		}

		if(verbose == true) print("setting %s (%d) to %d\n", VideoAttributes[va], previous_value, new_value);

		VPic.colour = new_value & 0xffff;
		if (ioctl(camfd, VIDIOCSPICT, &VPic) < 0) return false;
		return true;		
	}  
#else
	return false;
#endif
}

int VideoGrabberV4L::getGamma()
{
#ifdef HAVE_V4L
  	if((status != VIDEOSOURCE_OK) && (status != VIDEOSOURCE_OPEN)) return -1;

	return VPic.whiteness;
#else
	return -1;
#endif
}

bool VideoGrabberV4L::setGamma(int new_value)
{
#ifdef HAVE_V4L
  	int va = 4;
    	int previous_value = getGamma();

  	if(previous_value < 0) return false;
  	else
	{		  
		if((new_value < MinMax_VideoAttributes[va].min_va) || (new_value > MinMax_VideoAttributes[va].max_va))
		{			 
			warning("VideoGrabberV4L: Video Attribute (%s) value %d out of range, setting to default (%d)", VideoAttributes[va], new_value, MinMax_VideoAttributes[va].default_va); 

			new_value = MinMax_VideoAttributes[va].default_va;			 
		}

		if(verbose == true) print("setting %s (%d) to %d\n", VideoAttributes[va], previous_value, new_value);

		VPic.whiteness = new_value & 0xffff;
		if (ioctl(camfd, VIDIOCSPICT, &VPic) < 0) return false;
		return true;		
	}  
#else
	return false;
#endif
}


VideoGrabberV4L::~VideoGrabberV4L()
{
  	Flush();
}

}

