#ifndef VIDEO_GRABBER_V4L_INC
#define VIDEO_GRABBER_V4L_INC

#include "VideoGrabber.h"

//
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

//
#include <sys/types.h>

#ifdef HAVE_V4L
#include <linux/videodev.h>
#include "pwc-ioctl.h"
#endif

namespace Torch {

/** This class is designed to handle a V4L buffer
 
    @author Sebastien Marcel (marcel@idiap.ch) 
    @version 2.0
    \Date
    @since 1.0
*/
class FrameV4L: public VirtualPixmap
{
  	friend class VideoGrabberV4L;
  
private:
	//
	unsigned char *data;  
 	
	//
	int width, height; 

	//
  	int bytesperline_in;
  	int bytesperline_out;
  	
	//
  	int input_video_format;
	int output_video_format;
  
	//
  	bool bIsUsed;

	//
	bool verbose;

public:
  	///
  	FrameV4L(int width_, int height_, int input_video_format_, int output_video_format_, bool verbose_ = false);

	///
  	virtual ~FrameV4L();

	///
  	virtual void copyBuffer(unsigned char *);
  
	///
  	virtual unsigned int getwidth() const;

	///
  	virtual unsigned int getheight() const;
 
	///
	const int bytes_per_line() const { return 0; };
	
	///
	virtual const unsigned char *getpixmap() const;
};

/** This class is designed to grab images using V4L

    @author Sebastien Marcel (marcel@idiap.ch) 
    @version 2.0
    \Date
    @since 1.0
*/
class VideoGrabberV4L: public VideoGrabber
{
private:
   	//
  	FrameV4L **Frame;

	//
	int current_frame;
  
	//
  	int input_video_format;
  	int output_video_format;
  	int v4l_native_video_format;

	//
  	void Flush();

	// Interface to V4L data
	
	//
	char devname[40];

	//
	int camfd;

	//
#ifdef HAVE_V4L
	struct video_capability vcap;
	struct video_channel vchan;
	struct video_picture VPic;
	struct video_mbuf VMBuf;
	struct video_mmap vm;
#endif
	//
	int vid_io_buffer_size;
	int vid_io_image_size;
	unsigned char *vid_io_buffer;
	int *vid_io_offsets;
	int n_buffers;
	int video_channel;
	int frame_rate;
	int current_buffer;

	//
	bool validated;

	//-----
	
	//
	int Sync();

	//
	int Capture(int the_buffer_);

public:
	/** creates a video grabber for a V4L device 
	
	    @param devname_ is the name of the device (for instance "/dev/video0"
	    @param width_ is the #width# of the image to capture
	    @param height_ is the #height# of the image to capture
	    @param output_video_format_ is the output format to convert to
	    @param input_video_format_ is the native input format to capture
	    @param n_buffers_ is the number of buffers to use (auto if -1)
	    @param video_channel_ is the number of the video channel (S-Video, Composite1, Composite2, ...)
	    @param frame_rate_ is the framerate to use during the capture
	*/
  	VideoGrabberV4L(const char *devname_,
			int width_, int height_, 
			const char *output_video_format_ = "RGB24",
			const char *input_video_format_ = "YUV420P",
			int n_buffers_ = -1,
			int video_channel_ = 0,
			int frame_rate_ = 15,
			bool verbose_ = false);
  
	/// destructor
  	virtual ~VideoGrabberV4L();

	///
  	virtual bool initVideoSource();  

	///
  	virtual bool openVideoSource();

	///
  	virtual bool closeVideoSource();

	///
  	virtual VirtualPixmap* getframe();

	//-----
	
	///
  	virtual int getBrightness();
  	virtual bool setBrightness(int);

	///
  	virtual int getContrast();
  	virtual bool setContrast(int);
 
	///
  	virtual int getHue();
  	virtual bool setHue(int);
 
	///
  	virtual int getColor();
  	virtual bool setColor(int);

	///
  	virtual int getGamma();
  	virtual bool setGamma(int);
};

}

#endif
